// Copyright (c) Meta Platforms, Inc. and affiliates.

static void *opcode_targets[256] = {
    &&TARGET_CACHE,
    &&TARGET_POP_TOP,
    &&TARGET_PUSH_NULL,
    &&TARGET_INTERPRETER_EXIT,
    &&TARGET_END_FOR,
    &&TARGET_END_SEND,
    &&TARGET_BINARY_OP_ADD_FLOAT,
    &&TARGET_BINARY_OP_ADD_INT,
    &&TARGET_BINARY_OP_ADD_UNICODE,
    &&TARGET_NOP,
    &&TARGET_BINARY_OP_INPLACE_ADD_UNICODE,
    &&TARGET_UNARY_NEGATIVE,
    &&TARGET_UNARY_NOT,
    &&TARGET_BINARY_OP_MULTIPLY_FLOAT,
    &&TARGET_BINARY_OP_MULTIPLY_INT,
    &&TARGET_UNARY_INVERT,
    &&TARGET_BINARY_OP_SUBTRACT_FLOAT,
    &&TARGET_RESERVED,
    &&TARGET_BINARY_OP_SUBTRACT_INT,
    &&TARGET_BINARY_SUBSCR_DICT,
    &&TARGET_BINARY_SUBSCR_GETITEM,
    &&TARGET_BINARY_SUBSCR_LIST_INT,
    &&TARGET_BINARY_SUBSCR_TUPLE_INT,
    &&TARGET_CALL_PY_EXACT_ARGS,
    &&TARGET_CALL_PY_WITH_DEFAULTS,
    &&TARGET_BINARY_SUBSCR,
    &&TARGET_BINARY_SLICE,
    &&TARGET_STORE_SLICE,
    &&TARGET_CALL_BOUND_METHOD_EXACT_ARGS,
    &&TARGET_CALL_BUILTIN_CLASS,
    &&TARGET_GET_LEN,
    &&TARGET_MATCH_MAPPING,
    &&TARGET_MATCH_SEQUENCE,
    &&TARGET_MATCH_KEYS,
    &&TARGET_CALL_BUILTIN_FAST_WITH_KEYWORDS,
    &&TARGET_PUSH_EXC_INFO,
    &&TARGET_CHECK_EXC_MATCH,
    &&TARGET_CHECK_EG_MATCH,
    &&TARGET_CALL_METHOD_DESCRIPTOR_FAST_WITH_KEYWORDS,
    &&TARGET_CALL_NO_KW_BUILTIN_FAST,
    &&TARGET_CALL_NO_KW_BUILTIN_O,
    &&TARGET_CALL_NO_KW_ISINSTANCE,
    &&TARGET_CALL_NO_KW_LEN,
    &&TARGET_CALL_NO_KW_LIST_APPEND,
    &&TARGET_CALL_NO_KW_METHOD_DESCRIPTOR_FAST,
    &&TARGET_CALL_NO_KW_METHOD_DESCRIPTOR_NOARGS,
    &&TARGET_CALL_NO_KW_METHOD_DESCRIPTOR_O,
    &&TARGET_CALL_NO_KW_STR_1,
    &&TARGET_CALL_NO_KW_TUPLE_1,
    &&TARGET_WITH_EXCEPT_START,
    &&TARGET_GET_AITER,
    &&TARGET_GET_ANEXT,
    &&TARGET_BEFORE_ASYNC_WITH,
    &&TARGET_BEFORE_WITH,
    &&TARGET_END_ASYNC_FOR,
    &&TARGET_CLEANUP_THROW,
    &&TARGET_CALL_NO_KW_TYPE_1,
    &&TARGET_COMPARE_OP_FLOAT,
    &&TARGET_COMPARE_OP_INT,
    &&TARGET_COMPARE_OP_STR,
    &&TARGET_STORE_SUBSCR,
    &&TARGET_DELETE_SUBSCR,
    &&TARGET_FOR_ITER_LIST,
    &&TARGET_FOR_ITER_TUPLE,
    &&TARGET_FOR_ITER_RANGE,
    &&TARGET_FOR_ITER_GEN,
    &&TARGET_LOAD_SUPER_ATTR_ATTR,
    &&TARGET_LOAD_SUPER_ATTR_METHOD,
    &&TARGET_GET_ITER,
    &&TARGET_GET_YIELD_FROM_ITER,
    &&TARGET_LOAD_ATTR_CLASS,
    &&TARGET_LOAD_BUILD_CLASS,
    &&TARGET_LOAD_ATTR_GETATTRIBUTE_OVERRIDDEN,
    &&TARGET_LOAD_ATTR_INSTANCE_VALUE,
    &&TARGET_LOAD_ASSERTION_ERROR,
    &&TARGET_RETURN_GENERATOR,
    &&TARGET_LOAD_ATTR_MODULE,
    &&TARGET_LOAD_ATTR_PROPERTY,
    &&TARGET_LOAD_ATTR_SLOT,
    &&TARGET_LOAD_ATTR_WITH_HINT,
    &&TARGET_LOAD_ATTR_METHOD_LAZY_DICT,
    &&TARGET_LOAD_ATTR_METHOD_NO_DICT,
    &&TARGET_LOAD_ATTR_METHOD_WITH_VALUES,
    &&TARGET_RETURN_VALUE,
    &&TARGET_LOAD_CONST__LOAD_FAST,
    &&TARGET_SETUP_ANNOTATIONS,
    &&TARGET_LOAD_FAST__LOAD_CONST,
    &&TARGET_LOAD_LOCALS,
    &&TARGET_LOAD_FAST__LOAD_FAST,
    &&TARGET_POP_EXCEPT,
    &&TARGET_STORE_NAME,
    &&TARGET_DELETE_NAME,
    &&TARGET_UNPACK_SEQUENCE,
    &&TARGET_FOR_ITER,
    &&TARGET_UNPACK_EX,
    &&TARGET_STORE_ATTR,
    &&TARGET_DELETE_ATTR,
    &&TARGET_STORE_GLOBAL,
    &&TARGET_DELETE_GLOBAL,
    &&TARGET_SWAP,
    &&TARGET_LOAD_CONST,
    &&TARGET_LOAD_NAME,
    &&TARGET_BUILD_TUPLE,
    &&TARGET_BUILD_LIST,
    &&TARGET_BUILD_SET,
    &&TARGET_BUILD_MAP,
    &&TARGET_LOAD_ATTR,
    &&TARGET_COMPARE_OP,
    &&TARGET_EAGER_IMPORT_NAME,
    &&TARGET_IMPORT_FROM,
    &&TARGET_JUMP_FORWARD,
    &&TARGET_LOAD_GLOBAL_BUILTIN,
    &&TARGET_LOAD_GLOBAL_MODULE,
    &&TARGET_STORE_ATTR_INSTANCE_VALUE,
    &&TARGET_POP_JUMP_IF_FALSE,
    &&TARGET_POP_JUMP_IF_TRUE,
    &&TARGET_LOAD_GLOBAL,
    &&TARGET_IS_OP,
    &&TARGET_CONTAINS_OP,
    &&TARGET_RERAISE,
    &&TARGET_COPY,
    &&TARGET_RETURN_CONST,
    &&TARGET_BINARY_OP,
    &&TARGET_SEND,
    &&TARGET_LOAD_FAST,
    &&TARGET_STORE_FAST,
    &&TARGET_DELETE_FAST,
    &&TARGET_LOAD_FAST_CHECK,
    &&TARGET_POP_JUMP_IF_NOT_NONE,
    &&TARGET_POP_JUMP_IF_NONE,
    &&TARGET_RAISE_VARARGS,
    &&TARGET_GET_AWAITABLE,
    &&TARGET_MAKE_FUNCTION,
    &&TARGET_BUILD_SLICE,
    &&TARGET_JUMP_BACKWARD_NO_INTERRUPT,
    &&TARGET_MAKE_CELL,
    &&TARGET_LOAD_CLOSURE,
    &&TARGET_LOAD_DEREF,
    &&TARGET_STORE_DEREF,
    &&TARGET_DELETE_DEREF,
    &&TARGET_JUMP_BACKWARD,
    &&TARGET_LOAD_SUPER_ATTR,
    &&TARGET_CALL_FUNCTION_EX,
    &&TARGET_LOAD_FAST_AND_CLEAR,
    &&TARGET_EXTENDED_ARG,
    &&TARGET_LIST_APPEND,
    &&TARGET_SET_ADD,
    &&TARGET_MAP_ADD,
    &&TARGET_STORE_ATTR_SLOT,
    &&TARGET_COPY_FREE_VARS,
    &&TARGET_YIELD_VALUE,
    &&TARGET_RESUME,
    &&TARGET_MATCH_CLASS,
    &&TARGET_STORE_ATTR_WITH_HINT,
    &&TARGET_STORE_FAST__LOAD_FAST,
    &&TARGET_FORMAT_VALUE,
    &&TARGET_BUILD_CONST_KEY_MAP,
    &&TARGET_BUILD_STRING,
    &&TARGET_STORE_FAST__STORE_FAST,
    &&TARGET_STORE_SUBSCR_DICT,
    &&TARGET_STORE_SUBSCR_LIST_INT,
    &&TARGET_UNPACK_SEQUENCE_LIST,
    &&TARGET_LIST_EXTEND,
    &&TARGET_SET_UPDATE,
    &&TARGET_DICT_MERGE,
    &&TARGET_DICT_UPDATE,
    &&TARGET_UNPACK_SEQUENCE_TUPLE,
    &&TARGET_UNPACK_SEQUENCE_TWO_TUPLE,
    &&TARGET_SEND_GEN,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&TARGET_CALL,
    &&TARGET_KW_NAMES,
    &&TARGET_CALL_INTRINSIC_1,
    &&TARGET_CALL_INTRINSIC_2,
    &&TARGET_LOAD_FROM_DICT_OR_GLOBALS,
    &&TARGET_LOAD_FROM_DICT_OR_DEREF,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&TARGET_IMPORT_NAME,
    &&_unknown_opcode,
    &&TARGET_INVOKE_METHOD,
    &&TARGET_LOAD_FIELD,
    &&TARGET_LOAD_OBJ_FIELD,
    &&TARGET_LOAD_PRIMITIVE_FIELD,
    &&TARGET_STORE_FIELD,
    &&TARGET_STORE_OBJ_FIELD,
    &&TARGET_STORE_PRIMITIVE_FIELD,
    &&TARGET_BUILD_CHECKED_LIST,
    &&TARGET_BUILD_CHECKED_LIST_CACHED,
    &&TARGET_LOAD_TYPE,
    &&TARGET_CAST,
    &&TARGET_CAST_CACHED,
    &&TARGET_LOAD_LOCAL,
    &&TARGET_STORE_LOCAL,
    &&TARGET_STORE_LOCAL_CACHED,
    &&TARGET_PRIMITIVE_BOX,
    &&TARGET_POP_JUMP_IF_ZERO,
    &&TARGET_POP_JUMP_IF_NONZERO,
    &&TARGET_PRIMITIVE_UNBOX,
    &&TARGET_PRIMITIVE_BINARY_OP,
    &&TARGET_PRIMITIVE_UNARY_OP,
    &&TARGET_PRIMITIVE_COMPARE_OP,
    &&TARGET_LOAD_ITERABLE_ARG,
    &&TARGET_LOAD_MAPPING_ARG,
    &&TARGET_INVOKE_FUNCTION,
    &&TARGET_INVOKE_FUNCTION_CACHED,
    &&TARGET_INVOKE_INDIRECT_CACHED,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&TARGET_FAST_LEN,
    &&TARGET_CONVERT_PRIMITIVE,
    &&TARGET_INVOKE_NATIVE,
    &&TARGET_LOAD_CLASS,
    &&TARGET_BUILD_CHECKED_MAP,
    &&TARGET_BUILD_CHECKED_MAP_CACHED,
    &&TARGET_SEQUENCE_GET,
    &&TARGET_SEQUENCE_SET,
    &&TARGET_LIST_DEL,
    &&TARGET_REFINE_TYPE,
    &&TARGET_PRIMITIVE_LOAD_CONST,
    &&TARGET_RETURN_PRIMITIVE,
    &&TARGET_TP_ALLOC,
    &&TARGET_TP_ALLOC_CACHED,
    &&TARGET_LOAD_METHOD_STATIC,
    &&_unknown_opcode,
    &&TARGET_LOAD_METHOD_STATIC_CACHED,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&_unknown_opcode,
    &&TARGET_INSTRUMENTED_LOAD_SUPER_ATTR,
    &&TARGET_INSTRUMENTED_POP_JUMP_IF_NONE,
    &&TARGET_INSTRUMENTED_POP_JUMP_IF_NOT_NONE,
    &&TARGET_INSTRUMENTED_RESUME,
    &&TARGET_INSTRUMENTED_CALL,
    &&TARGET_INSTRUMENTED_RETURN_VALUE,
    &&TARGET_INSTRUMENTED_YIELD_VALUE,
    &&TARGET_INSTRUMENTED_CALL_FUNCTION_EX,
    &&TARGET_INSTRUMENTED_JUMP_FORWARD,
    &&TARGET_INSTRUMENTED_JUMP_BACKWARD,
    &&TARGET_INSTRUMENTED_RETURN_CONST,
    &&TARGET_INSTRUMENTED_FOR_ITER,
    &&TARGET_INSTRUMENTED_POP_JUMP_IF_FALSE,
    &&TARGET_INSTRUMENTED_POP_JUMP_IF_TRUE,
    &&TARGET_INSTRUMENTED_END_FOR,
    &&TARGET_INSTRUMENTED_END_SEND,
    &&TARGET_INSTRUMENTED_INSTRUCTION,
    &&TARGET_INSTRUMENTED_LINE,
    &&_unknown_opcode
};
