# Copyright (c) Meta Platforms, Inc. and affiliates.
# pyre-strict

from _cinderx import StaticTypeError
from _static import (
    __build_cinder_class__,
    _clear_dlopen_cache,
    _clear_dlsym_cache,
    _property_missing_fget,
    _property_missing_fset,
    _sizeof_dlopen_cache,
    _sizeof_dlsym_cache,
    chkdict,
    chklist,
    FAST_LEN_ARRAY,
    FAST_LEN_DICT,
    FAST_LEN_INEXACT,
    FAST_LEN_LIST,
    FAST_LEN_SET,
    FAST_LEN_STR,
    FAST_LEN_TUPLE,
    init_subclass,
    install_sp_audit_hook,
    is_static_callable,
    is_static_module,
    is_type_static,
    lookup_native_symbol,
    make_context_decorator_wrapper,
    make_recreate_cm,
    PRIM_OP_ADD_DBL,
    PRIM_OP_ADD_INT,
    PRIM_OP_AND_INT,
    PRIM_OP_DIV_DBL,
    PRIM_OP_DIV_INT,
    PRIM_OP_DIV_UN_INT,
    PRIM_OP_EQ_DBL,
    PRIM_OP_EQ_INT,
    PRIM_OP_GE_DBL,
    PRIM_OP_GE_INT,
    PRIM_OP_GE_UN_INT,
    PRIM_OP_GT_DBL,
    PRIM_OP_GT_INT,
    PRIM_OP_GT_UN_INT,
    PRIM_OP_INV_INT,
    PRIM_OP_LE_DBL,
    PRIM_OP_LE_INT,
    PRIM_OP_LE_UN_INT,
    PRIM_OP_LSHIFT_INT,
    PRIM_OP_LT_DBL,
    PRIM_OP_LT_INT,
    PRIM_OP_LT_UN_INT,
    PRIM_OP_MOD_DBL,
    PRIM_OP_MOD_INT,
    PRIM_OP_MOD_UN_INT,
    PRIM_OP_MUL_DBL,
    PRIM_OP_MUL_INT,
    PRIM_OP_NE_DBL,
    PRIM_OP_NE_INT,
    PRIM_OP_NEG_DBL,
    PRIM_OP_NEG_INT,
    PRIM_OP_NOT_INT,
    PRIM_OP_OR_INT,
    PRIM_OP_POW_DBL,
    PRIM_OP_POW_INT,
    PRIM_OP_POW_UN_INT,
    PRIM_OP_RSHIFT_INT,
    PRIM_OP_RSHIFT_UN_INT,
    PRIM_OP_SUB_DBL,
    PRIM_OP_SUB_INT,
    PRIM_OP_XOR_INT,
    rand,
    RAND_MAX,
    resolve_primitive_descr,
    SEQ_ARRAY_INT64,
    SEQ_CHECKED_LIST,
    SEQ_LIST,
    SEQ_LIST_INEXACT,
    SEQ_REPEAT_INEXACT_NUM,
    SEQ_REPEAT_INEXACT_SEQ,
    SEQ_REPEAT_PRIMITIVE_NUM,
    SEQ_REPEAT_REVERSED,
    SEQ_SUBSCR_UNCHECKED,
    SEQ_TUPLE,
    set_type_code,
    set_type_final,
    set_type_static,
    set_type_static_final,
    staticarray,
    TYPED_ARRAY,
    TYPED_BOOL,
    TYPED_CHAR,
    TYPED_DOUBLE,
    TYPED_INT16,
    TYPED_INT32,
    TYPED_INT64,
    TYPED_INT8,
    TYPED_INT_16BIT,
    TYPED_INT_32BIT,
    TYPED_INT_64BIT,
    TYPED_OBJECT,
    TYPED_SINGLE,
    TYPED_UINT16,
    TYPED_UINT32,
    TYPED_UINT64,
    TYPED_UINT8,
)
